# -*- coding: utf-8 -*-

#*******************************************************************************
#
# Organics
# ---------------------------------------------------------
#
#
# (C) 2020 by TEA SISTEMI S.p.a. (iacopo.borsi@tea-group.com) as Consultant of Scuola Superiore Sant'Anna
#
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU General Public License as published by  *
# *   the Free Software Foundation; either version 2 of the License, or     *
# *   (at your option) any later version.                                   *
# *                                                                         *
# ******************************************************************************
import os
import shutil,sqlite3
import pandas as pd
from matplotlib.dates import DateFormatter

# for PyQGIS 3.x
from qgis.PyQt.QtCore import Qt, QSettings, QVariant, QDate, QDateTime, QTime, QFile
from qgis.PyQt.QtWidgets import QDialog, QAction, QFileDialog, QMessageBox, QDialogButtonBox
from qgis.PyQt import QtGui, uic
from qgis.core import *
from qgis.core import QgsVectorLayer, QgsCoordinateReferenceSystem, QgsProcessingFeatureSourceDefinition, QgsVectorFileWriter
from qgis.gui import QgsCheckableComboBox
import processing
#
from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5 import QtWidgets
from PyQt5 import QtGui, uic
from qgis.PyQt.QtCore import Qt, QSettings, QVariant, QDate, QDateTime, QTime, QFile
#
import numpy as np
from scipy import special
import matplotlib.pyplot as plt
from . import resources
#

#
FORM_CLASS, _ = uic.loadUiType(os.path.join( os.path.dirname(__file__), 'organics_dialog.ui') )

class OrganicsDockWidget(QDialog, FORM_CLASS):
    def __init__(self, iface):
        QDialog.__init__(self)
        self.iface = iface
        self.setupUi(self)
        self.setWindowIcon(QtGui.QIcon(os.path.join( os.path.dirname(__file__), 'icon.png')))
        #When clicking the browse_button button, recall the method loadReadcsv
        self.txtInputFile.fileChanged.connect(self.loadReadcsv)
        #When clicking the run_button button, recall the method run
        self.run_button.clicked.connect(self.run)
        self.buttonBox.button(QDialogButtonBox.Cancel).clicked.connect(self.reject)
        self.buttonBoxPlot.button(QDialogButtonBox.Ok).clicked.connect(self.plotResults)
        self.buttonBoxPlot.button(QDialogButtonBox.Cancel).clicked.connect(self.reject)
        #
        self.cmbLine.setFilters(QgsMapLayerProxyModel.LineLayer)
        self.cmbLine.setShowCrs(True)
        #
        self.cmbResults.setFilters(QgsMapLayerProxyModel.PointLayer)
        self.cmbResults.setShowCrs(True)
        self.cmbResults.currentIndexChanged.connect(self.reloadFields)
        #
        self.reloadFields()
        # Inlcude images.
        pixmapBase = QPixmap(os.path.join( os.path.dirname(__file__), 'GREVISLIN.gif'))
        self.lbl_image_1.setPixmap(pixmapBase)
        self.lbl_image_2.setPixmap(pixmapBase)


    def tr(self, message):
        """Get the translation for a string using Qt translation API.

        We implement this ourselves since we do not inherit QObject.

        :param message: String for translation.
        :type message: str, QString

        :returns: Translated version of message.
        :rtype: QString
        """
        # noinspection PyTypeChecker,PyArgumentList,PyCallByClass
        return QCoreApplication.translate('OrganicsDockWidget', message)

    def reject(self):
        #
        QDialog.reject(self)


    def loadReadcsv(self):
        #getOpenFileName is a method of the QFileDialog class which allows to open a dialog window with
        #the heading "Select input csv file".
        #The argument "" makes this dialog window to open always at the same path, while '*.csv' allows
        #to filter csv files.
        #The getOpenFileName method returns the csv file inputfile and res, which allows to check if
        #I clicked Open or not
        #I use self.inputfile (and not just inputfile) because I will need this variable in the
        #retrieve*Data functions below
        self.inputfile = self.txtInputFile.filePath()

        # --- Opening the csv input file just selected
        csvData = pd.read_csv(self.inputfile, sep = ';', parse_dates = True, infer_datetime_format = True)
        self.csvData = csvData
        headers=list(csvData.keys())
        #Inserting the string 'Select a field...' at the beginning of
        headers.insert(0,self.tr('Select a field...'))

        #Filling the combo boxes time_field, conc_inlet and v_field
        #of the Dialog with headers of the csv input file (addItems is a method
        #of the QComboBox class which provides a list of options in a combo box)
        self.time_field.clear()
        self.time_field.addItems(headers)
        self.conc_inlet.clear()
        self.conc_inlet.addItems(headers)
        self.v_field.clear()
        self.v_field.addItems(headers)

        #Once the text in the time_field combo box is changed,
        #recall the retrieveTimeData function
        self.time_field.currentIndexChanged.connect(self.getDataFields)
        #Once the text in the conc_inlet combo box is changed,
        #recall the retrieveTimeData function
        self.conc_inlet.currentIndexChanged.connect(self.getDataFields)
        #Once the text in the v_field combo box is changed,
        #recall the retrieveVelocityData function
        self.v_field.currentIndexChanged.connect(self.getDataFields)


    def getDataFields(self):
        # Get data:
        self.t_field=self.time_field.currentText()
        self.cinlet_field=self.conc_inlet.currentText()
        self.velocity_field = self.v_field.currentText()

    def pde_solver(self, x,t,v,k,E,C0, tau):
        g = np.sqrt(1+4*(k*E)/(v*v)) # Gamma coefficient
        if tau is None or t < tau:
            a1 = (np.exp(v*x*(1-g)/(2*E)))*special.erfc((x-v*t*g)/(2*np.sqrt(E*t)))
            a2 = (np.exp(v*x*(1+g)/(2*E)))*special.erfc((x+v*t*g)/(2*np.sqrt(E*t)))
            c = C0*(a1+a2)/2
        else:
            a = (np.exp(v*x*(1-g)/(2*E)))
            b = (np.exp(v*x*(1+g)/(2*E)))
            a1 = special.erfc((x-v*t*g)/(2*np.sqrt(E*t))) - special.erfc((x-v*(t-tau)*g)/(2*np.sqrt(E*(t-tau))))
            b1 = special.erfc((x+v*t*g)/(2*np.sqrt(E*t))) - special.erfc((x+v*(t-tau)*g)/(2*np.sqrt(E*(t-tau))))
            c = C0*(a*a1+b*b1)/2
        return c

    def pde_solver_no_pulse(self, x,t,v,k,E,C0):
        g = np.sqrt(1+4*(k*E)/(v*v)) # Gamma coefficient
        a1 = (np.exp(v*x*(1-g)/(2*E)))*special.erfc((x-v*t*g)/(2*np.sqrt(E*t)))
        a2 = (np.exp(v*x*(1+g)/(2*E)))*special.erfc((x+v*t*g)/(2*np.sqrt(E*t)))
        c = C0*(a1+a2)/2
        return c

    #The run method allows to simulate organics' concentration at the outlet, based on the
    #measured concentration at the inlet and on the travel time of the sample collected at the inlet until the outlet
    def run(self):
        # get data
        self.getDataFields()
        # dispersion coefficient:
        E = self.spinDiffusion.value()
        # degradation rate :
        k = self.spinDegradation.value()


        #Checking the content of the combo box time_field of the Dialog
        if self.t_field== 'Select a field...':
            msg = self.tr('Select a field for Measurement time')
            QtWidgets.QMessageBox.warning(None, 'Error!', '%s' % (msg))
            return

        #Checking the content of the combo box conc_inlet of the Dialog
        if self.cinlet_field=='Select a field...':
            msg =  self.tr('Select a field for Concentration measured at the inlet')
            QtWidgets.QMessageBox.warning(None, 'Error!', '%s' % (msg))
            return

        #Checking the content of the combo box v_field of the Dialog
        if self.velocity_field=='Select a field...':
            msg =  self.tr('Select a field for Average velocity of the stream at the inlet')
            QtWidgets.QMessageBox.warning(None, 'Error!', '%s' % (msg))
            return

        # -------------
        # Split line in sections according to the length given in input
        line = self.cmbLine.currentLayer()
        sfCount = line.selectedFeatureCount()
        if sfCount <= 0:
            line.selectByIds([k.id() for k in line.getFeatures()])

        distance = int(self.spinDistance.value())
        myresult = processing.run("qgis:pointsalonglines",{ 'DISTANCE' : distance, 'END_OFFSET' : 0, 'INPUT' : QgsProcessingFeatureSourceDefinition(line.id(), True), 'OUTPUT' : 'memory:', 'START_OFFSET' : 0 })
        points = myresult['OUTPUT']
        # get the max distance
        distList = [float(f['distance']) for f in points.getFeatures()]
        distList = np.array(distList)
        maxDist = np.max(distList)
        print('ditanza massima = ', maxDist)


        # add time field and DeltaT
        newfields = [ QgsField('date', QVariant.String), QgsField('dt', QVariant.Double), QgsField('conc', QVariant.Double) ]
        points.dataProvider().addAttributes(newfields)
        points.updateFields()


        # Manage the vector x ... vector of discrete distance between sections
        x = np.array([0.0])
        distList = []





        # per prendere dato con filtro:
        # aa['Diclofenac (ng/l)'][aa['Date and Time'] == '2018-05-28 18:32:00' ]

        # manage time: set as ordered array, to be sure!
        date_time_ordered = np.sort(self.csvData[self.t_field])

        # first loop on features, to initialize time series
        #  Start editing to write record
        points.startEditing()
        changedAttributesDict = {}


        totalCountTime = len(date_time_ordered)
        date_before = date_time_ordered[0]
        if totalCountTime > 1:
            mydate_string = date_time_ordered[1]
            init_date = pd.to_datetime(date_before)
            mydate = pd.to_datetime(mydate_string)
            deltaT = (mydate - pd.to_datetime(date_before)).total_seconds()
        else:
            # max time (in s.) of simulation after the intial time (in hours from GUI)
            #
            init_date = pd.to_datetime(date_before)
            v = float(self.csvData[self.velocity_field][self.csvData[self.t_field]==date_before])
            Tmax = maxDist/v
            deltaT = Tmax


        #
        time_step = int(self.spinTs.value()*60.0)  # time step in seconds (in minutes from GUI)
        #
        nsteps = int(deltaT/time_step)

        print('primo giro, deltaT, time_step : ',deltaT,time_step)

        if nsteps > 1:
            dtVector = np.array(range(0,int(deltaT),time_step))
        else:
            dtVector = np.array([deltaT])


        self.progressBar.setValue(1)
        itBar = 0
        totalCountPoints = points.featureCount()
        totalProgressTime = totalCountPoints*totalCountTime
        # initialize new fields with 1st initial condition
        for ft in points.getFeatures():
            changedAttributes = {}
            x = (float(ft['distance']))
            #
            v = float(self.csvData[self.velocity_field][self.csvData[self.t_field]==date_before])
            C0 = float(self.csvData[self.cinlet_field][self.csvData[self.t_field]==date_before])
            # calculate concentration using amalytical solution:
            if x == 0.0:
                myconc = C0
            else:
                myconc = 0.0

            # update values
            fieldId_dt = points.dataProvider().fieldNameIndex('dt')
            changedAttributes[fieldId_dt] = 0.0
            #
            fieldId_date = points.dataProvider().fieldNameIndex('date')
            thisDate = init_date
            changedAttributes[fieldId_date] = str(thisDate.strftime('%Y-%m-%d %H:%M:%S'))
            #
            fieldId_c= points.dataProvider().fieldNameIndex('conc')
            changedAttributes[fieldId_c] = float(myconc)
            # update dictionary
            changedAttributesDict[ft.id()] = changedAttributes
            #
            itBar += 100*1/totalProgressTime
            self.progressBar.setValue(itBar)

        points.dataProvider().changeAttributeValues(changedAttributesDict)
        points.commitChanges()
        #
        masterFeaturesList = [ft for ft in points.getFeatures()]
        # proceed for other time steps of 1st stress period (if any):
        if nsteps > 1:
            newFeatureList = []
            for ft in masterFeaturesList:
                for ts in dtVector[1:]: # t=0 already processed: start from 1 !
                    if ts <= 0:
                        continue
                    # add copy of the feature
                    feat = QgsFeature(points.fields())
                    feat.id = totalCountTime + ft.id() + 1
                    # feat.setGeometry(new_geometry)
                    feat.setGeometry(ft.geometry())
                    # update values
                    x = (float(ft['distance']))
                    #
                    v = float(self.csvData[self.velocity_field][self.csvData[self.t_field]==date_before])
                    C0 = float(self.csvData[self.cinlet_field][self.csvData[self.t_field]==date_before])
                    #
                    # calculate concentration using analytical solution:
                    if x == 0.0:
                        myconc = C0
                    else:
                        myconc = self.pde_solver(x,ts,v,k,E,C0,deltaT)
                        # smooth to 0:
                        if myconc <= 1e-6:
                            myconc = 0.0
                    #
                    feat.setAttribute('distance', x)
                    feat.setAttribute('dt', float(ts))
                    feat.setAttribute('conc',float(myconc))
                    thisDate = init_date + pd.Timedelta(seconds=ts)
                    feat.setAttribute('date', str(thisDate.strftime('%Y-%m-%d %H:%M:%S')))
                    #
                    newFeatureList.append(feat)
            #
            points.startEditing()
            (res, outFeats) = points.dataProvider().addFeatures(newFeatureList)
            points.commitChanges()
        # update total time of simulation
        Tsim = deltaT
        TsimList = [Tsim]
        T0List = [0]
        ConcList = [C0]
        vList = [v]


        # start looping on time (on stress periods):
        if totalCountTime > 1:
            #
            for i in range(1,len(date_time_ordered)):
                # all but one stress periods:
                if i < len(date_time_ordered)-1:
                    mydate_string = date_time_ordered[i+1]
                    date_before = date_time_ordered[i]
                    print('Inizia stress period {}'.format(1+i))
                    print('data vecchia', date_before, 'data nuova', mydate_string)
                    mydate = pd.to_datetime(mydate_string)
                    init_date = pd.to_datetime(date_before)
                    deltaT = (mydate - pd.to_datetime(date_before)).total_seconds()
                    print('durata intervallo stress period', deltaT)
                    print('tempo iniziale', TsimList[-1])
                    #dtVector = np.array(range(int(Tsim),int(deltaT+Tsim),time_step))
                    dtVector = np.array(range(0,int(deltaT),time_step))
                    nsteps = len(dtVector)
                    print('vettore tempo', dtVector)
                    vnew = float(self.csvData[self.velocity_field][self.csvData[self.t_field]==date_before])
                    Cnew = float(self.csvData[self.cinlet_field][self.csvData[self.t_field]==date_before])
                # the last StressPeriod:
                else:
                    date_before = date_time_ordered[-1]
                    init_date = pd.to_datetime(date_before)
                    # calculate Tmax:
                    vnew = float(self.csvData[self.velocity_field][self.csvData[self.t_field]==date_before])
                    Cnew = float(self.csvData[self.cinlet_field][self.csvData[self.t_field]==date_before])
                    Tmax = maxDist/vnew
                    deltaT = Tmax
                    #dtVector = np.array(range(int(Tsim),int(deltaT+Tsim),time_step))
                    dtVector = np.array(range(0,int(deltaT),time_step))
                    print('data iniziale ultimo stress period', date_before)
                    print('data iniziale ultimo stress period in PD dataframe', init_date)
                    nsteps = len(dtVector)

                # update list of values:
                ConcList.append(Cnew)
                T0List.append(TsimList[-1])
                TsimList.append(TsimList[-1] + deltaT)
                vList.append(vnew)
                newFeatureList = []
                print('Liste memoria: Conc, Tsim, T0, v0')
                print(ConcList)
                print(TsimList)
                print(T0List)
                print(vList)

                for jf, ft in enumerate(masterFeaturesList):
                    #
                    for ts in dtVector:
                        if i == 0 and ts == 0:
                            continue
                        # add copy of the feature
                        feat = QgsFeature(points.fields())
                        feat.id = totalCountTime + 1
                        # feat.setGeometry(new_geometry)
                        feat.setGeometry(ft.geometry())
                        # update values
                        x = (float(ft['distance']))
                        #
                        tnow = T0List[-1] + ts  # absolute time
                        #
                        # calculate concentration using analytical solution:

                        myconc = 0.0
                        # build the solution using the super-position principle
                        for kj in range(0,len(TsimList)):
                            t = tnow - T0List[kj]
                            Cnew = ConcList[kj]
                            v = vList[kj]
                            if kj == 0:
                                Cold = 0
                            else:
                                Cold = ConcList[kj-1]
                            # solution for this init condition
                            tau = TsimList[kj] - T0List[kj]
                            # ctemp = self.pde_solver(x,t,v,k,E,Cnew - Cold, tau)
                            ctemp = self.pde_solver_no_pulse(x,t,v,k,E,Cnew - Cold)
                            #
                            # update the sum
                            myconc = myconc + ctemp

                            if float(x) == 100.0:
                                print('------------- step ', kj, ' ------------------------')
                                print('t assoluto  = {}'.format(tnow))
                                print('t0 assoluto = {}, conc totale = {}'.format(T0List[kj],myconc))
                                print('tempo t relativo = {}, tau = {}'.format(t,tau))
                                print('v0 ora = {}, conc diff  = {}'.format(v,Cnew-Cold))
                                print('C0 ora = {}, conc parziale = {}'.format(Cnew,ctemp))

                            #
                        # smooth to 0:
                        if myconc <= 1e-6:
                            myconc = 0.0
                        if x == 0:
                            myconc = ConcList[-1]

                        feat.setAttribute('distance', x)
                        #
                        feat.setAttribute('dt', float(tnow) )
                        feat.setAttribute('conc',float(myconc))
                        thisDate = init_date + pd.Timedelta(seconds=(ts))
                        if jf == 0:
                            print('Sono a x = 0')
                            print('tnow in sec. = ', tnow, 'conc = ', myconc, 'data iniziale = ', init_date, 'data nuova = ', thisDate)
                        #
                        feat.setAttribute('date', str(thisDate.strftime('%Y-%m-%d %H:%M:%S')))
                        #
                        newFeatureList.append(feat)
                    # update progress bar
                    #

                #
                itBar += 100*1/totalProgressTime
                self.progressBar.setValue(itBar)
                #
                points.startEditing()
                (res, outFeats) = points.dataProvider().addFeatures(newFeatureList)
                points.commitChanges()
                #
            #

        if self.txtOutputFile.filePath() == '' :
            # Comment the following line to avoid showing the points in Legend:
            QgsProject.instance().addMapLayer(points)
        else:
            # save in shape file:
            outputName = os.path.basename(str(self.txtOutputFile.filePath())).split('.')[0]
            outputLayer = QgsVectorFileWriter.writeAsVectorFormat(points,str(self.txtOutputFile.filePath()),"UTF-8",driverName="ESRI Shapefile")
            vlayer = QgsVectorLayer(str(self.txtOutputFile.filePath()), outputName, "ogr")
            QgsProject.instance().addMapLayer(vlayer)


        msg =  self.tr('Process completed!')
        QtWidgets.QMessageBox.information(None, 'Done!', '%s' % (msg))



    def getResultsFromShape(self):
        resultsLayer = self.cmbResults.currentLayer()
        pathResult = os.path.dirname(resultsLayer.dataProvider().dataSourceUri())

        try:
            nameCsv = os.path.basename(resultsLayer.dataProvider().dataSourceUri()).split('.')[0]
            fullPathCsv = os.path.join(pathResult,nameCsv + '.csv')
            csvlayer = QgsVectorFileWriter.writeAsVectorFormat(resultsLayer,fullPathCsv,"UTF-8",driverName="CSV")
        except:
            nameCsv = 'output.csv'
            fullPathCsv = os.path.join(os.getcwd(),nameCsv)
            print(fullPathCsv)
            csvlayer = QgsVectorFileWriter.writeAsVectorFormat(resultsLayer,fullPathCsv,"UTF-8",driverName="CSV")
        #

        self.outputData = pd.read_csv(fullPathCsv, sep = ',', parse_dates = True, infer_datetime_format = True)

    #
    def reloadFields(self):
        try:
            self.getResultsFromShape()
            tempArray = np.unique(self.outputData['distance'])
            # make unique
            self.cmbFieldX.clear()
            itemsList = [str(d) for d in tempArray]
            self.cmbFieldX.addItems(itemsList)

            # and for Date Field:
            tempArray = np.unique(self.outputData['date'])
            # make unique
            self.cmbFieldDate.clear()
            itemsList = [str(d) for d in tempArray]
            self.cmbFieldDate.addItems(itemsList)
        except:
            self.cmbFieldX.addItems(['Select a Shape Layer of Results'])
            self.cmbFieldDate.addItems(['Select a Shape Layer of Results'])



    def plotResults(self):

        # get units of measure for Concentrations
        units = str(self.txtUnits.text())

        if self.radioX.isChecked():
            x = np.float64(self.cmbFieldX.currentText())
            myseries = self.outputData[self.outputData['distance'] == x].sort_values(by=['date'])
            myseries['conc']=pd.to_numeric(myseries['conc'])
            #
            formatter = DateFormatter('%Y-%m %H:%M')
            fig, ax = plt.subplots()
            ax.plot(myseries['date'],myseries['conc'])

            ax.set(xlabel = self.tr('Time'), ylabel = self.tr('Concentration [{}]').format(units), title = self.tr('Evolution of concentration at {} m from inlet').format(x))
            ax.set_xticks(myseries['date'])
            ax.grid()
            fig.autofmt_xdate()
            plt.show()
            return
        elif self.radioDate.isChecked():
            t = str(self.cmbFieldDate.currentText())
            myseries = self.outputData[self.outputData['date'] == t].sort_values(by=['distance'])
            myseries['conc']=pd.to_numeric(myseries['conc'])
            #myx = self.outputData['distance'][self.outputData['date'] == t]
            #plt.plot(myx,myseries)
            fig, ax = plt.subplots()
            ax.plot(myseries['distance'],myseries['conc'])
            ax.set(xlabel = self.tr('Distance (m)'), ylabel = self.tr('Concentration [{}]').format(units), title = self.tr('Concentration at {}').format(t))
            ax.set_xticks(myseries['distance'])
            plt.grid()
            plt.show()
            return
        else:
            targetPoints = self.cmbResults.currentLayer()
            if targetPoints.selectedFeatureCount() > 0:
                xList = []
                for ft in targetPoints.selectedFeatures():
                    x = float(ft['distance'])
                    xList.append(x)
                # make unique
                xList = np.unique(xList)
                for j, xj in enumerate(xList):
                    myseries = self.outputData['conc'][self.outputData['distance'] == xj]
                    mytime = self.outputData['date'][self.outputData['distance'] == xj]
                    plt.plot(mytime,myseries, label = '{} m from inlet'.format(xj))
                #
                plt.xticks(rotation=60)
                plt.ylabel(self.tr('Concentration [{}]').format(units))
                plt.title(self.tr('Evolution of concentration'))
                plt.grid()
                plt.legend(loc = 'best').set_draggable(True)
                plt.show()
                return
            else:
                msg =  self.tr('Please, select at least 1 Point in Layer of Results!')
                QtWidgets.QMessageBox.warning(None, 'Error!', '%s' % (msg))
                return








